/*
 * dnode.c - HP-UX node functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dnode.c,v 1.2 97/10/23 12:23:30 abe Exp $";
#endif


#include "lsof.h"
#include <sys/inode.h>

#if	HPUXV>=900
_PROTOTYPE(static void enter_nma,(char *b));
_PROTOTYPE(static char islocked,(struct locklist *lp));
#endif	/* HPUXV>=900 */

_PROTOTYPE(static int readinode,(struct inode *ia, struct inode *i));


#if	HPUXV>=900
/*
 * enter_nma() - enter NAME column addition
 */

static void
enter_nma(b)
	char *b;			/* addition buffer */
{
	char *cp;
	MALLOC_S len;

	if (Lf->nma)
		return;
	if ((len = (MALLOC_S)(strlen(b) + 1)) < 2)
		return;
	if ((cp = (char *)malloc(len)) == (char *)NULL)
		return;
	(void) strcpy(cp, b);
	Lf->nma = cp;
}


/*
 * islocked() - is node locked?
 */

static char
islocked(lp)
	struct locklist *lp;		/* local locklist struct pointer */
{
	int l;
	struct locklist ll;
	KA_T llf, llp;

	if (!(llf = (KA_T)lp))
	    return(' ');
	llp = llf;
/*
 * Search the locklist chain for this process.
 */
	do {
	    if (kread(llp, (char *)&ll, sizeof(ll)))
		return(' ');

#if	!defined(L_REMOTE)
#define	L_REMOTE	0x1		/* from HP-UX 9.01 */
#endif	/* !defined(L_REMOTE) */

# if	HPUXV<1010
	    if (ll.ll_flags & L_REMOTE || ll.ll_proc != (struct proc *)Kpa)
# else	/* HPUXV>=1010 */
	    if (ll.ll_flags & L_REMOTE || (caddr_t)ll.ll_kthreadp != Ktp)
# endif	/* HPUXV<1010 */

		continue;
	    if (ll.ll_start == 0 && ll.ll_end == 0x7fffffff)
		l = 1;
	    else
		l = 0;
	    if (ll.ll_type == F_WRLCK)
		return(l ? 'W' : 'w');
	    else if (ll.ll_type == F_RDLCK)
		return(l ? 'R' : 'r');
	    return(' ');
	}

# if	HPUXV<1010
	while ((llp = (KA_T)ll.ll_link) && llp != llf);
# else	/* HPUXV>=1010 */
	while ((llp = (KA_T)ll.ll_fwd) && llp != llf);
# endif	/* HPUXV<1010 */

	return(' ');
}
#endif	/* HPUXV>=900 */


/*
 * process_node() - process vnode
 */

void
process_node(va)
	caddr_t va;			/* vnode kernel space address */
{
	static unsigned long avops;
	static unsigned long cvops;
	dev_t dev;
	char dev_ch[32];
	static int ft = 1;
	static unsigned long fvops;
	struct inode i;
	int ins = 0;
	struct vfs *la = NULL;
	static unsigned long nvops;
	static unsigned long pvops;
	struct rnode r;
	int rns = 0;
	static unsigned long svops;
	char *ty;
	enum vtype type;
	static unsigned long uvops;
	static struct vnode *v = (struct vnode *)NULL;
	struct l_vfs *vfs;
	static unsigned long vvops;

#if	defined(HAS_AFS)
	static int afs = 0;		/* AFS test status: -1 = no AFS
					 *		     0 = not tested
					 *		     1 = AFS present */
	struct afsnode an;
#endif	/* defined(HAS_AFS) */

#if	HPUXV>=900
	char fb[128];
	int fns = 0;
	int rp, sz, wp;
	struct vnode rv;
	struct snode s;
#endif	/* HPUXV>=900 */

#if	HPUXV>=1000
	struct cdnode c;
	struct fifonode f;
	struct vattr vat;
	int vats = 0;
#endif	/* HPUXV>=1000 */

/*
 * Do first-time only operations.
 */
	if (ft) {
		if (get_Nl_value("avops", Drive_Nl, &avops) < 0)
			avops = (unsigned long)0;
		if (get_Nl_value("cvops", Drive_Nl, &cvops) < 0)
			cvops = (unsigned long)0;
		if (get_Nl_value("fvops", Drive_Nl, &fvops) < 0)
			fvops = (unsigned long)0;
		if (get_Nl_value("nvops", Drive_Nl, &nvops) < 0)
			nvops = (unsigned long)0;
		if (get_Nl_value("pvops", Drive_Nl, &pvops) < 0)
			pvops = (unsigned long)0;
		if (get_Nl_value("svops", Drive_Nl, &svops) < 0)
			svops = (unsigned long)0;
		if (get_Nl_value("uvops", Drive_Nl, &uvops) < 0)
			uvops = (unsigned long)0;
		if (get_Nl_value("vvops", Drive_Nl, &vvops) < 0)
			vvops = (unsigned long)0;
		ft = 0;
	}
/*
 * Read the vnode.
 */
	if ( ! va) {
		enter_nm("no vnode address");
		return;
	}
	if (!v) {

	/*
	 * Allocate space for the vnode or AFS vcache structure.
	 */

#if	defined(HAS_AFS)
		v = alloc_vcache();
#else	/* !defined(HAS_AFS) */
		v = (struct vnode *)malloc(sizeof(struct vnode));
#endif	/* defined(HAS_AFS) */

		if (!v) {
			(void) fprintf(stderr, "%s: can't allocate %s space\n",
				Pn,

#if	defined(HAS_AFS)
				"vcache"
#else	/* !defined(HAS_AFS) */
				"vnode"
#endif	/* defined(HAS_AFS) */

				);
			Exit(1);
		}
	}
	if (readvnode((caddr_t)va, v)) {
                enter_nm(Namech);
                return;
        }

#if	defined(HASNCACHE)
	Lf->na = (unsigned long)va;
#endif	/* defined(HASNCACHE) */

/*
 * Determine the vnode type.
 */
	if (uvops && uvops == (unsigned long)v->v_op)
		Ntype = N_REGLR;
	else if (nvops && nvops == (unsigned long)v->v_op)
		Ntype = N_NFS;

#if	defined(HASVXFS)
	else if (vvops && vvops == (unsigned long)v->v_op)
		Ntype = N_VXFS;
#endif	/* defined(HASVXFS) */

#if	HPUXV>=1000
	else if (cvops && cvops == (unsigned long)v->v_op)
		Ntype = N_CDFS;
	else if (fvops && fvops == (unsigned long)v->v_op)
		Ntype = N_FIFO;
	else if (pvops && pvops == (unsigned long)v->v_op)
		Ntype = N_PIPE;
	else if (svops && svops == (unsigned long)v->v_op)
		Ntype = N_SPEC;
#else	/* HPUXV<1000 */
	else if (v->v_type == VFIFO)
		Ntype = N_FIFO;
#endif	/* HPUXV<1000 */

#if	defined(HAS_AFS)
	/*
	 * Caution: this AFS test should be the last one.
	 */
	
	else if (avops) {
		if (avops == (unsigned long)v->v_op)
			Ntype = N_AFS;
		else {

unknown_v_op:
			(void) sprintf(Namech,
				"unknown file system type; v_op: %#x",
				v->v_op);
			enter_nm(Namech);
			return;
		}
	} else if (v->v_data || !v->v_vfsp)
		goto unknown_v_op;
	else {
		switch (afs) {
		case -1:
			goto unknown_v_op;
		case 0:
			if (!hasAFS(v)) {
				afs = -1;
				goto unknown_v_op;
			}
			afs = 1;
			Ntype = N_AFS;
			break;
		case 1:
			if (v->v_vfsp == AFSVfsp)
				Ntype =  N_AFS;
			else
				goto unknown_v_op;
		}
	}
#else	/* !defined(HAS_AFS) */
	else {
		(void) sprintf(Namech, "unknown file system type; v_op: %#x",
			v->v_op);
		enter_nm(Namech);
		return;
	}
#endif	/* defined(HAS_AFS) */

#if	HPUXV<900
	if (v->v_shlockc || v->v_exlockc) {
		if (v->v_shlockc && v->v_exlockc)
			Lf->lock = 'u';
		else if (v->v_shlockc)
			Lf->lock = 'R';
		else
			Lf->lock = 'W';
	}
#else	/* HPUXV>900 */
# if	HPUXV>=1000
/*
 * Determine lock type.
 */
	Lf->lock = islocked(v->v_locklist);
# endif	/* HPUXV>=1000 */
#endif	/* HPUXV<900 */

/*
 * Establish the local virtual file system structure.
 */
	if ( ! v->v_vfsp)
		vfs = NULL;
	else if ((vfs = readvfs(la, v)) == NULL) {
                (void) sprintf(Namech, "can't read vfs for %#x at %#x", va,
			v->v_vfsp);
                enter_nm(Namech);
		return;
	}
/*
 * Read the cdnode, fifonode, inode, rnode, snode, or vache struct.
 */
	switch (Ntype) {

#if	defined(HAS_AFS)
	case N_AFS:
		if (readafsnode(va, v, &an))
			return;
		break;
#endif	/* defined(HAS_AFS) */

#if	defined(HASVXFS)
	case N_VXFS:
		if (!v->v_data || read_vxnode(v, vfs, &dev)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read vx_inode (%#x)",
				va, v->v_data);
			enter_nm(Namech);
			return;
		}
		break;
#endif	/* defined(HASVXFS) */

#if	HPUXV>=1000
	case N_CDFS:
		if (!v->v_data
		||  kread((KA_T)v->v_data, (char *)&c, sizeof(c))) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read cdnode (%#x)",
				va, v->v_data);
			enter_nm(Namech);
			return;
		}
		break;
	case N_FIFO:
	case N_PIPE:
		if (!v->v_data
		||  kread((KA_T)v->v_data, (char *)&f, sizeof(f))) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read fifonode (%#x)",
				va, v->v_data);
			enter_nm(Namech);
			return;
		}
		fns = 1;
		if (f.fn_vap
		&&  kread((KA_T)f.fn_vap, (char *)&vat, sizeof(vat)) == 0)
			vats = 1;
		break;
#endif	/* HPUXV>=1000 */

	case N_NFS:
		if (!v->v_data || readrnode((caddr_t)v->v_data, &r)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read rnode (%#x)",
				va, v->v_data);
			enter_nm(Namech);
			return;
		}
		rns = 1;
		break;

#if	HPUXV>=1000
	case N_SPEC:
		if (v->v_type == VCHR) {
			if (!v->v_data || readsnode(v->v_data, &s)) {
				(void) sprintf(Namech,
					"vnode at %#x: can't read snode(%#x)",
					va, v->v_data);
				enter_nm(Namech);
				return;
			}
			if ( ! s.s_realvp
			||  readvnode((caddr_t)s.s_realvp, &rv)) {
				(void) sprintf(Namech,
				    "snode at %#x: can't read real vnode (%#x)",
				    v->v_data, s.s_realvp);
				enter_nm(Namech);
				return;
			}
			if ( ! rv.v_data
			||  readinode((struct inode *)rv.v_data, &i)) {
				(void) sprintf(Namech,
					"snode at %#x: can't read inode (%#x)",
					v->v_data, rv.v_data);
				enter_nm(Namech);
				return;
			}
			ins = 1;
			break;
		}
		if (!v->v_data || readinode((struct inode *)v->v_data, &i)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read inode (%#x)",
				va, v->v_data);
			enter_nm(Namech);
			return;
		}
		ins = 1;
		break;
#endif	/* HPUXV>=1000 */

#if	HPUXV>=900 && HPUXV<1000
	case N_FIFO:
		if (v->v_fstype == VNFS_FIFO) {
		    if ( ! v->v_data
		    ||   readsnode((caddr_t)v->v_data, &s)) {
			(void) sprintf(Namech,
			    "fifonode at %#x: can't read snode (%#x)",
			    va, v->v_data);
			enter_nm(Namech);
			return;
		    }
		    if ( ! s.s_realvp
		    ||  readvnode((caddr_t)s.s_realvp, &rv)) {
			(void) sprintf(Namech,
			    "fifonode at %#x: can't read real vnode (%#x)",
			    v->v_data, s.s_realvp);
			enter_nm(Namech);
			return;
		    }
		    if ( ! rv.v_data
		    ||   readrnode((caddr_t)rv.v_data, &r)) {
			(void) sprintf(Namech,
			    "fifonode at %#x: can't read real rnode (%#x)",
			    v->v_data, s.s_realvp);
			enter_nm(Namech);
			return;
		    }
		    rns = 1;
		    break;
		}
		/* fall through */
#endif	/* HPUXV>=900 && HPUXV<1000 */

	case N_REGLR:
	default:
		if (v->v_type != VBLK) {
		    if (!v->v_data
		    ||  readinode((struct inode *)v->v_data, &i)) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read inode (%#x)",
				va, v->v_data);
			enter_nm(Namech);
			return;
		    }
		    ins = 1;
		}

#if	HPUXV>=900 && HPUXV<1000
		if (v->v_type == VFIFO)
			Ntype = N_FIFO;
#endif	/* HPUXV>=900 && HPUXV<1000 */

	}

#if	HPUXV>=900 && HPUXV<1000
	Lf->lock = islocked(i.i_locklist);
#endif	/* HPUXV>=900 && HPUXV<1000 */

/*
 * Get device and type for printing.
 */
	switch (Ntype) {

#if	defined(HAS_AFS)
	case N_AFS:
		dev = an.dev;
		break;
#endif	/* defined(HAS_AFS) */

	case N_NFS:
		dev = vfs ? vfs->dev : 0;
		break;

#if	HPUXV>=1000
	case N_CDFS:
		dev = c.cd_dev;
		break;
	case N_FIFO:
	case N_PIPE:
		if (vfs && vfs->fsname)
			dev = vfs->dev;
		else {
			(void) sprintf(dev_ch, "0x%08x", va);
			enter_dev_ch(dev_ch);
		}
		break;
#endif	/* _HPUX>=1000 */

#if	defined(HASVXFS)
	case N_VXFS:
		/* obtained via read_vxnode */
		break;
#endif	/* defined(HASVXFS) */

	case N_SPEC:
	default:

#if	HPUXV>=800
		if (v->v_type == VCHR || v->v_type == N_BLK)
			dev = v->v_rdev;
		else if (vfs && vfs->fsname)
			dev = vfs->dev;
		else
			dev = i.i_dev;
#else	/* HPUXV<800 */
		dev = (v->v_type == VCHR || v->v_type == VBLK) ? v->v_rdev
							       : i.i_dev;
#endif	/* HPUXV>=800 */

	}

	type = v->v_type;
/*
 * Obtain the inode number.
 */
	switch (Ntype) {

#if	defined(HAS_AFS)
	case N_AFS:
		if (an.ino_st) {
			Lf->inode = an.inode;
			Lf->inp_ty = 1;
		}
		break;
#endif	/* defined(HAS_AFS) */

	case N_NFS:
		Lf->inode = (unsigned long)r.r_nfsattr.na_nodeid;
		Lf->inp_ty = 1;
		break;

#if	HPUXV>=1000
	case N_CDFS:
		Lf->inode = (unsigned long)c.cd_num;
		break;
	case N_FIFO:
	case N_PIPE:
		if (vats) {
			Lf->inode = (unsigned long)vat.va_nodeid;
			Lf->inp_ty = 1;
		} else {
			Lf->inode = (unsigned long)v->v_nodeid;
			Lf->inp_ty = 1;
		}
		break;
#endif	/* HPUXV>=1000 */

#if	defined(HASVXFS)
	case N_VXFS:
		/* set in read_vxnode() */
		break;
#endif	/* defined(HASVXFS) */

#if	HPUXV<1000
	case N_FIFO:

# if	HPUXV>=900
		if (rns) {
			Lf->inode = (unsigned long)r.r_nfsattr.na_nodeid;
			Lf->inp_ty = 1;
			break;
		}
# endif	/* HPUXV>=900 */
		/* fall through */

#endif	/* HPUXV<1000 */

	case N_REGLR:
	case N_SPEC:
		if (ins) {
			Lf->inode = (unsigned long)i.i_number;
			Lf->inp_ty = 1;
		}
	}
/*
 * Obtain the file size.
 */
	if (Foffset)
		Lf->off_def = 1;
	else {
		switch (Ntype) {

#if	defined(HAS_AFS)
		case N_AFS:
			Lf->sz = (SZOFFTYPE)an.size;
			Lf->sz_def = 1;
			break;
#endif	/* defined(HAS_AFS) */


#if	HPUXV>=1000
		case N_CDFS:
			Lf->sz = (SZOFFTYPE)c.cd_cdc.cdc_size;
			Lf->sz_def = 1;
			break;
		case N_PIPE:
			if (vats) {
				Lf->sz = (SZOFFTYPE)vat.va_size;
				Lf->sz_def = 1;
			}
			break;
#endif	/* HPUXV>=1000 */

#if	HPUXV>=900
		case N_FIFO:

# if	HPUXV<1000
			if (ins) {
			    rp = i.i_frptr;
			    sz = (int)i.i_fifosize;
			    wp = i.i_fwptr;
			} else if (rns)
			    Lf->sz = (SZOFFTYPE)r.r_nfsattr.na_size;
# else	HPUXV>=1000 */
			if (fns) {
			    rp = f.fn_rptr;
			    sz = f.fn_size;
			    wp = f.fn_wptr;
			}
# endif	/* HPUXV<1000 */

			if (Fsize || (Lf->access != 'r' && Lf->access != 'w')) {
			    if (fns || ins) {
				(void) sprintf(fb, "rd=%#x; wr=%#x", rp, wp);
				(void) enter_nma(fb);
			    }
			    if (fns || ins || rns) {
				Lf->sz = (SZOFFTYPE)sz;
				Lf->sz_def = 1;
			    }
			    break;
			}
			if (fns || ins) {
			    Lf->off = (unsigned long)((Lf->access == 'r') ? rp
									  : wp);
			    (void) sprintf(fb, "%s=%#x",
				(Lf->access == 'r') ? "rd" : "wr",
				(Lf->access == 'r') ?  rp  :  wp);
			    (void) enter_nma(fb);
			}
			Lf->off_def = 1;
			break;
#endif	/* HPUXV>=900 */
		case N_NFS:
			Lf->sz = (SZOFFTYPE)r.r_nfsattr.na_size;
			Lf->sz_def = 1;
			break;

#if	defined(HASVXFS)
		case N_VXFS:
			/* set in read_vxnode() */
			break;
#endif	/* defined(HASVXFS) */

		case N_SPEC:
		case N_REGLR:
			if (type == VREG || type == VDIR) {
				Lf->sz = (SZOFFTYPE)i.i_size;
				Lf->sz_def = 1;
			} else if ((type == VCHR || type == VBLK) && !Fsize)
				Lf->off_def = 1;
			break;
		}
	}
/*
 * Record an NFS file selection.
 */
	if (Ntype == N_NFS && Fnfs)
		Lf->sf |= SELNFS;
/*
 * Save the file system names.
 */
	if (vfs) {
		Lf->fsdir = vfs->dir;
		Lf->fsdev = vfs->fsname;

#if	defined(HASFSINO)
		Lf->fs_ino = vfs->fs_ino;
#endif	/* defined(HASFSINO) */

	}
/*
 * Format the vnode type, and possibly the device name.
 */
	switch (type) {

	case VNON:
		ty ="VNON";
		break;
	case VREG:
	case VDIR:
		ty = (type == VREG) ? "VREG" : "VDIR";
		Lf->dev_def = 1;
		Lf->dev = dev;
		break;
	case VBLK:
		ty = "VBLK";
		Lf->dev_def = 1;
		Lf->dev = dev;
		Ntype = N_BLK;
		break;
	case VCHR:
		ty = "VCHR";
		Lf->dev = dev;
		Lf->dev_def = 1;
		Ntype = N_CHR;
		break;
	case VLNK:
		ty = "VLNK";
		break;

#if	defined(VSOCK)
	case VSOCK:
		ty = "SOCK";
		break;
#endif

	case VBAD:
		ty = "VBAD";
		break;
	case VFIFO:
		switch (Ntype) {

#if	HPUXV>=1000
		case N_FIFO:
			if (vfs && vfs->fsname) {
				Lf->dev = dev;
				Lf->dev_def = 1;
			}
			ty = "FIFO";
			break;
		case N_PIPE:
			if (vfs && vfs->fsname) {
				Lf->dev = dev;
				Lf->dev_def = 1;
			}
			ty = "PIPE";
			break;
#endif	/* HPUXV>=1000 */

		default:
			Lf->dev = dev;
			Lf->dev_def = 1;
			ty = "FIFO";
		}
		break;
	default:
		if (type > 9999)
			(void) sprintf(Lf->type, "*%03d", type % 1000);
		else
			(void) sprintf(Lf->type, "%4d", type);
		(void) strcpy(Namech, "unknown type");
		ty = NULL;
	}
	if (ty)
		(void) strcpy(Lf->type, ty);
	Lf->ntype = Ntype;

#if	defined(HASBLKDEV)
/*
 * If this is a VBLK file and it's missing an inode number, try to
 * supply one.
 */
	if (Lf->inp_ty == 0 && type == VBLK && Lf->dev_def)
		find_bl_ino();
#endif	/* defined(HASBLKDEV) */

/*
 * If this is a VCHR file and it's missing an inode number, try to
 * supply one.
 */
	if (Lf->inp_ty == 0 && type == VCHR && Lf->dev_def)
		find_ch_ino();
/*
 * Test for specified file.
 */
	if (Sfile && is_file_named(NULL, type))
		Lf->sf |= SELNM;
/*
 * Enter name characters.
 */
	if (Namech[0])
		enter_nm(Namech);
}



/*
 * readinode() - read inode
 */

static int
readinode(ia, i)
	struct inode *ia;		/* inode kernel address */
	struct inode *i;		/* inode buffer */
{
	if (kread((KA_T)ia, (char *)i, sizeof(struct inode))) {
		(void) sprintf(Namech, "can't read inode at %#x", ia);
		return(1);
	}
	return(0);
}
